#include "vmath.h"
#include "fpmath.h"
#include <stdio.h>

// Optimized SSE ray-box test.
// Note how our float3 class means we get all the benefit of SSE optimization,
// but our code is just as clear as the regular version below.
bool intersectRayBox(float3 rayOrg, float3 invDir, float3 bbmin, float3 bbmax, float &hitT)
{
	float3 d0 = (bbmin - rayOrg) * invDir;
	float3 d1 = (bbmax - rayOrg) * invDir;

	float3 v0 = min(d0, d1);
	float3 v1 = max(d0, d1);

	float tmin = hmax(v0);
	float tmax = hmin(v1);

	bool hit = (tmax >= 0) && (tmax >= tmin) && (tmin <= hitT);
	if (hit)
		hitT = tmin;
	return hit;
}

// Regular floating-point math ray-box test.
bool intersectRayBoxFP(const vec3 &rayOrg, const vec3 &invDir, const vec3 &bbmin, const vec3 &bbmax, float &hitT)
{
	vec3 d0 = (bbmin - rayOrg) * invDir;
	vec3 d1 = (bbmax - rayOrg) * invDir;

	vec3 v0 = min(d0, d1);
	vec3 v1 = max(d0, d1);

	float tmin = hmax(v0);
	float tmax = hmin(v1);

	bool hit = (tmax >= 0) && (tmax >= tmin) && (tmin <= hitT);
	if (hit)
		hitT = tmin;
	return hit;
}

// Test data.
float3 testorg(-10, 0, 0);
float3 testdir(1, 0, 0);
float3 testbbmin = float3(-4, -4, -4);
float3 testbbmax = float3(+4, +4, +4);
vec3 testorgFP = { -10, 0, 0 };
vec3 testdirFP = { 1, 0, 0 };
vec3 testbbminFP = { -4, -4, -4 };
vec3 testbbmaxFP = { +4, +4, +4 };

void testSSE()
{
	float hitT = FLT_MAX;
	bool hit = intersectRayBox(testorg, float3(1, 1, 1) / testdir, testbbmin, testbbmax, hitT);
	printf("hit %i at t=%f\n", hit, hitT);
}

void testFP()
{
	float hitTFP = FLT_MAX;
	vec3 oneFP = { 1, 1, 1 };
	bool hitFP = intersectRayBoxFP(testorgFP, oneFP / testdirFP, testbbminFP, testbbmaxFP, hitTFP);
	printf("hit %i at t=%f\n", hitFP, hitTFP);
}

int __cdecl main()
{
	testSSE();
	testFP();
	return 0;
}
